/** A mapping from K to V. Null keys are not allowed. */
interface MapI<K,V> {
  
  /** create a new empty map of the same type. */
  MapI<K,V> newMap();
  
  /** @return true if the map is empty (contains no key-value pairs). */
  boolean isEmpty();
  
  /** Set this map to the empty map. */
  void clear();
  
  /** @return the value for key if there is an entry (key,value) in the map; null otherwise. */
  V get(K key);
  
  /** Puts the entry (key, value) in the map, replacing the old entry for key if one existed.
    * @return the old value for key or null if no old value existed in the map. */
  V put(K key, V value);
  
  /* Deletes the entry for key.
   * @return the value of the removed entry if such an entry existed, null otherwise. */
  V remove(K key);
  
  /* return true iff an entry for key exists in the map. */
  public boolean containsKey(K key);
  
  /** Returns the number of entries (keys) in this map. */
  int size();
}

class MyHashMap<K,V> {
  
  static int INIT_CAPACITY = 1;
  static double MAX_LOAD = 1.;
  
  Node<K,V> buckets[];
  
  int capacity;
  
  double size; // size is exact because both must be less than 2*32.
  
  MyHashMap() { this(INIT_CAPACITY); }
  
  @SuppressWarnings("unchecked")
  MyHashMap(int c) { 
    capacity = c;
    buckets = (Node<K,V>[]) new Node[capacity];  // No first class generics!  Ugh.
    size = 0.;
  }
  
  MyHashMap newMap() { return new MyHashMap<K,V>(); }
  
  boolean isEmpty() { return size == 0.; }
  
  @SuppressWarnings("unchecked")
  void clear() {
    capacity = INIT_CAPACITY;
    buckets = (Node<K,V> []) new Node[capacity];
    size = 0.;
  }

  V get(K key) {
    final int index = key.hashCode() % capacity;
    Node<K,V> current = buckets[index];
    while (current != null) {
      if (key.equals(current.key)) return current.value;
      current = current.next;
    }
    return null;
  }
  
  V put(K key, V value) {

    final int index = key.hashCode() % capacity;
    final Node<K,V> head = buckets[index];
    Node<K,V> current = head;
    while (current != null) {
      if (key.equals(current.key)) break;
      current = current.next;
    }
    V oldValue;
    if (current != null) {
      oldValue = current.value;
      current.value = value;
      return oldValue;
    }
 
    /* key is not present in table; insert <key, value> at head of bucket list. */
    buckets[index] = new Node<K,V>(key, value, head);  // bucket may/may not be empty
    size++;
    double load = size/capacity;
    if (load > MAX_LOAD) expand();
    return null;
  }

  V remove(K key) {
    final int index = key.hashCode() % capacity;
    final Node<K,V> head = buckets[index];
    if (head == null) return null;
    if (head.key.equals(key)) {  /* Remove head node. */
      V oldValue = head.value;
      buckets[index] = head.next;
      size--;
      return oldValue;
    }
    /* bucket is non-empty. */
    Node<K,V> prev = head;
    Node<K,V> current = head.next;
      
    while (current != null) {
      if (key.equals(current.key)) break;
      prev = current;
      current = current.next;
    }

    if (current == null) return null; 
    
    /* Remove matching node. */
    V oldValue = current.value;
    prev.next = current.next;  // remove Node
    size--;
    return oldValue;
  }
  
  boolean containsKey(K key) { return get(key) != null; }
  
  int size() { return (int) size; }
  
  @SuppressWarnings("unchecked")
  private void expand() {
    capacity = 2 * capacity;
    Node<K,V>[] oldBuckets = buckets;
    size = 0;
    
    buckets = (Node<K,V>[]) new Node[capacity];  // No first class generics!  Ugh.
    
    for (Node<K,V> n: oldBuckets) {  // rehash entries in bucket n
      while (n != null) {   
        put(n.key, n.value);
        n = n.next;
      }
    }
  } 
  
  private static class Node<K,V> {
    K key;
    V value;
    Node<K,V> next;
    
    Node(K k, V v, Node<K,V> n) {
      key = k;
      value = v;
      next = n;
    }
  }
}
  