/** The interface for a Comp211 collection of objects of type E. */
interface EnumI<E> {
  
  /** Returns true if this enumeration is empty. */
  boolean isEmpty();
  /** Returns a ReadIteratorI for processing this enumeration. */
  ReadIteratorI<E> newIterator();
}

/** An iterator that traverses a collection of objects of type E (such as an EnumI<E>) without modifying it. */
interface ReadIteratorI<E> {
  /** Moves cursor to the first element of the collection.  Behavior unspecified on empty collecton. */
  E first();
  /** Moves cursor to the next element of the collection.  Behavior unspecified on empty collecton. */ 
  E next();
  /** Returns true if cursor is pointing to the last element of the collection. */
  boolean atEnd();
  /** Returns true if cursor is pointing to the first element of the collection. */
  boolean atStart();
  /** Returns the current element (pointed to by the cursor). 
    * @throws IteratorException if no such element exists.
    */
  E currentItem();
}

/** A mutable linear list of elements of type E that can be scanned from first (front) to last (rear).  
  * This interface can easily implemented as a singly linked list. */
interface ListI<E> extends EnumI<E> {
  /** Returns a new empty list. */
  ListI<E> newList();
  /** Returns the length of this list. */
  int length();
  /** Inherited method with narrowed output type. */
  BiListI<E> insertFront(E o);
  /** Inserts the object o at the rear of this list. */
  BiListI<E> insertRear(E o);
  /** Removes the object at the front of the list and returns it. */
  E remFront();
  /** Clears this;
    * @return this */
  BiList<E> clear();
  /** The inherited newIterator with a narrowed output type */
  IteratorI<E> newIterator();
}

/** An iterator that traverses a collection of objects of type E (such as an EnumI<E>) possibly modifying it. */
interface IteratorI<E> extends ReadIteratorI<E> {
  
  /** Destructively inserts the object o immediately before the current item.  If the list is empty, or the cursor 
    * is atEnd(), the node is inserted immediately before the end of the list. */
  void insert(E o);
  
  /** Removes the current item.
    * @throws an IteratorException if the cursor is atEnd(). */
  E remove();
}

/* A mutable list that can be scanned in either direction. */
interface BiListI<E> extends ListI<E> {
  
  /** The nherited newList with a narrowed output type. */
  BiListI<E> newList();
  
  /** Removes the last element of this list. 
    * @throws a BiListException if no such element exists. 
    */
  E remRear();
  
  /** The inherited newIterator with a narrowed output type. */
  BiIteratorI<E> newIterator(); 
}

/** An iterator that traverses a collection of objects of type E in either direction possibly modifying it. */
interface BiIteratorI<E> extends IteratorI<E> {
  /** Places the cursor on the last element of the collection.  The behavior is unspecified if the collection is empty. */
  E last(); 
  /** Places the cursor on the prevoius element of the collection.  The behavior is unspecified if the collection is empty. */
  E prev(); 
}

interface ILambda<A,B> {
  B apply(A host); 
}
/** Exception class for signalling errors in processing lists (implementing ListI) */
class ListException extends RuntimeException {
  ListException(String s) { super(s); }
}

/** Exception class for signalling errors in using an iterator (implementing IteratorI) */
class IteratorException extends RuntimeException {
  IteratorException(String s) { super(s); }
}

/** A circularly linked class implementing the BiListI interface. */
class BiList<E> implements BiListI<E> {
  
  /** The header node for the list. */
  private Node<E> head = new Node<E>();
  /** The length of this list. */
  private int length = 0;
  
  /* Relying on the default constructor for BiList. */
  
  public String toString(){
    StringBuffer s = new StringBuffer("(");
    for (IteratorI i = newIterator(); ! i.atEnd(); i.next()) {
      if (i.atStart()) s.append(i.currentItem().toString());
      else s.append(" " + i.currentItem().toString());
    }  
    return s.append(")").toString();
  }
 
  /* Implementations of BiListI methods.  See BiListI for contracts. */
  public boolean isEmpty() { return head.succ == head; } 
  public BiList<E> newList() { return new BiList<E>(); }
  public int length() { return length; }
  
  public BiList<E> clear() {
    head.pred = head;
    head.succ = head;
    length = 0;
    return this;
  }
  
  public <R> BiList<R> map(ILambda<E, R> f) {
    BiIteratorI<E> it = newIterator();
    BiList<R> result = new BiList<R>();
    while (! it.atEnd()) {  // atEnd() is true after going one next() beyond the first element
      E current = it.currentItem();
//      System.err.println("inserting " + f.apply(current) + " in " + result);
      result.insertRear(f.apply(current));
      it.next();
    }
    return result;
  }
  public BiList<E> insertFront(E o) { 
    Node<E> next = head.succ;
    Node<E> newNode = new Node<E>(o,head,next);  // allocate new Node
    // insert new Node
    head.succ = newNode;
    next.pred = newNode;
    length++;
    return this;
  }
  
  public BiList<E> insertRear(E o)  { 
    Node<E> prev = head.pred;
    Node<E> newNode = new Node<E>(o,prev,head);  // allocate new Node
    // insert new Node
    head.pred = newNode;
    prev.succ = newNode;
    length++;
    return this;
  }
  
  public E remFront() {
    if (isEmpty()) 
      throw new 
      BiListException("Attempt to remove first element of empty CircList");
    Node<E> next = head.succ;
    E remob = next.item;
    Node<E> newNext = next.succ;
    head.succ = newNext;
    newNext.pred = head;
    length--;
    
    return remob;
  }
  
  /* Virtual constructors for iterators; identical except for return type */
  public BiIteratorI<E> newIterator() { return new BiIterator(); }
  
  public E remRear() {
    if (isEmpty())
      throw new 
      BiListException("Attempt to remove last element of empty CircList");
    Node<E> last = head.pred;
    E remob = last.item;
    Node<E> lastPred = last.pred;
    head.pred = lastPred;
    lastPred.succ = head;
    length--;
    return remob;
  }
  
  boolean delete(E elt) {
    for (IteratorI<E> i = newIterator(); !i.atEnd(); i.next())
      if (i.currentItem().equals(elt)) {
      i.remove();
      return true;
    };
    return false;
  }
  
  /* Static nested Node class.  Embedded as nested class to hide access. */
  private static class Node<E> {
    E item;
    Node<E> pred,succ;
    
    Node(E i, Node<E> p, Node<E> s) {
      item = i;
      pred = p;
      succ = s;
    }
    
    Node() {  // allocate header
      item = null;
      pred = this;
      succ = this;
    }
  } 
  
  /** A true inner class that includes a link (BiList.this) to the enclosing BiList. */
  private class BiIterator implements BiIteratorI<E> {
    
    Node<E> current;
    
    BiIterator() {
      current = BiList.this.head.succ;  // current is first item (if one exists)
    }
    
    public E first() { 
      current = BiList.this.head.succ;  // current is first item (if one exists)
      return current.item;
    }
    
    public E last() { 
      current = BiList.this.head.pred;  // current is last item (if one exists)
      return current.item;
    }
    
    public E next() {
      current = current.succ;        // wraps around end
      return current.item;
    }
    
    public E prev() {
      current = current.pred;        // wraps around end
      return current.item;
    }
    
    public E currentItem() { 
      if (current == BiList.this.head) throw
        new IteratorException("No current element in " + BiList.this);
      return current.item; 
    }
    
    public boolean atEnd() { return current == BiList.this.head; }
   
    /** Returns true if CircList is non-empty and current is the first node in the list. */
    public boolean atStart() { return ! atEnd() && current == BiList.this.head.succ; }
    
    
    public void insert(E o) {
      
      // pre: true
      // post: Node containing o is inserted before current item, 
      //       current is unchanged
      
      Node<E> oldPred = current.pred;
      Node<E> newNode = new Node<E>(o, oldPred, current);  // allocate new node
      current.pred = newNode;                              // insert it
      oldPred.succ = newNode;
      BiList.this.length++;
      
    } 
    
    public E remove() {
      
      // pre:  current is valid
      // post: current becomes current.succ
      
      if (current == BiList.this.head) throw 
        new IteratorException("BiIterator.remove() applied at end of BiList " + BiList.this);
      Node<E> cPred = current.pred;
      Node<E> cSucc = current.succ;
      cPred.succ = cSucc;
      cSucc.pred = cPred;
      E remOb = current.item;
      current = cSucc;  
      BiList.this.length--;
      return remOb;
    } 
  }
  
  static class BiListException extends RuntimeException {
    BiListException(String s) { super(s); }
  }
}