/** A mapping from type K to type V. Null keys are not allowed. */
interface MapI<K,V> {
  
  /** create a new empty map of the same type MapI<K,V>. */
  MapI<K,V> newMap();
  
  /** @return true if the map is empty, i.e., contains no entries (key-value pairs). */
  boolean isEmpty();
  
  /** Set this map to the empty map. */
  void clear();
  
  /** @return the value for key if there is an entry (key,value) in the map; null otherwise. */
  V get(K key);
  
  /** Puts the entry (key, value) in the map, replacing the old entry for key if one existed.
    * @return the old value for key or null if no old value existed in the map. */
  V put(K key, V value);
  
  /* Deletes the entry for key.
   * @return the value of the removed entry if such an entry existed, null otherwise. */
  V remove(K key);
  
  /* return true iff an entry for key exists in the map. */
  public boolean containsKey(K key);
  
  /** Returns the number of entries (keys) in this map. */
  int size();
}

/** A map over a totally-ordered domain K. */
interface SortedMapI<K extends Comparable<K>, V> extends MapI<K, V> {
  K firstKey();
  K lastKey();
}

/** A mutable binary search tree of entries of type (K,V). */ 
class TreeMap<K extends Comparable<K>, V> implements SortedMapI<K,V> {
   
  /** Number of entries in the map. */
  private int size;
  
  /** Root of the binary search tree.  Weak form of state pattern: either empty or node. 
    * An empty tree is makeEmpty == new RefNode(null).  A node is RefNode(n) where n is a Node. */
  private RefNode<K,V> root;
  
  /* Constructors */
  TreeMap() { 
    root = makeEmpty();
    size = 0;
  }
  
  TreeMap(K key, V val) {
    root = makeLeaf(key, val);
    size = 1;
  }
  
  private TreeMap(RefNode<K,V> rn, int s) {
    root = rn;
    size = s;
  }
  
  /** Factory method for constructing empty trees. */
  private static <K extends Comparable<K>, V> RefNode<K,V> makeEmpty() { return new RefNode<K,V>(null); } 

  /** Factory method for constructing leaves. */
  private static <K extends Comparable<K>, V> RefNode<K,V> makeLeaf(K key, V val) { return new RefNode<K,V>(key, val); }
  
  /** create a new empty map of the same type. */
  public MapI<K,V> newMap() { return new TreeMap<K,V>(); }
  
  /** @return true if the map is empty (contains no key-value pairs). */
  public boolean isEmpty() { return root.isEmpty(); }
  
  /** Set this map to the empty map. */
  public void clear() { 
    root = makeEmpty();
    size = 0;
  }
  
  /** @return the value for key if there is an entry (key,value) in the map; null otherwise. */
  public V get(K key) { return find(key).value(); }  // returns null if find(key) is empty
  
  /** Puts the entry (key, value) in the map, replacing the old entry for key if one existed.
    * @return the old value for key or null if no old value existed in the map.
    */
  public V put(K k, V v) {
    // location is the matching node which may be empty
    RefNode<K,V> location = find(k);  // may return empty
    V oldValue = location.value();    // may return null
    if (location.isEmpty()) {  // key k does not yet exist in tree
      location.set(k,v);
      size++;
    }
    else location.setValue(v);  // key k already exists in tree
    return oldValue;
  }
  
  /* Deletes the entry for key if it exists. 
   * @return the value of the removed entry if such an entry existed, null otherwise. */
  public V remove(K key) {
    RefNode<K,V> location = find(key);  // may return root or empty location
    if (location.isEmpty()) return null;
    V oldVal =  location.remove(this);  // returns old value in location; decrements size
    return oldVal;
  } 
  
  /** @returns the matching RefNode, which is empty if ! containsKey(key) */
  private RefNode<K,V> find(K key) {
    RefNode<K,V> current = root;
    while (! current.isEmpty()) {
      int compareFlag = key.compareTo(current.key());
      if (compareFlag == 0) break;
      else if (compareFlag < 0) current = current.left();
      else current = current.right();
    }
    return current;
  }
  
  /** Requires: ! root.isEmpty()
    * @returns the first key in this. */
  public K firstKey() { 
    assert ! root.isEmpty();
    return root.findFirst().key(); 
  }
  
  /** Requires: ! root.isEmpty()
    * @returns the last key. */
  public K lastKey() {
    assert ! root.isEmpty();
    return root.findLast().key(); 
  }
    
  /* @return true iff an entry for key exists in the map. */
  public boolean containsKey(K key) { return get(key) != null; }
  
  /** @return the size of this map. */
  public int size() { return size; }
  
  public String toString() {
    if (isEmpty()) { return "{}"; }
    
    StringBuilder sb = new StringBuilder();
    toStringHelp(sb, root);
    
    /* sb has a leading space */
    sb.setCharAt(0, '{');
    sb.append('}');
    return sb.toString();
  }
  
  private void toStringHelp(StringBuilder sb, RefNode<K,V> n) {
    if (! n.isEmpty()) {
      toStringHelp(sb, n.left());
      sb.append(" [" + n.key() + "," + n.value() + "]");
      toStringHelp(sb, n.right());
    }
  }
  
  /** A reference class (box) representing an updateable cell that holds a Node. */
  private static class RefNode<K extends Comparable<K>,V> {
    Node<K,V> node;
    
    /** Constructors */
    RefNode() { this(null); }
    RefNode(K key, V val) { this(new Node<K,V> (key, val)); }
    
    private RefNode(Node<K,V> n) { node = n; }

    boolean isEmpty() { return node == null; }
    K key() { return node.key; }
    V value() {
      if (isEmpty()) return null;
      else return node.value; 
    }
    
    /** Requires: ! isEmpty().  
      * @return left subtree */
    RefNode<K,V> left() { 
      assert ! isEmpty();
      return node.left; 
    }
    
    /** Requires: ! isEmpty().
      * @return left subtree */
    RefNode<K,V> right() { 
      assert ! isEmpty();
      return node.right;
    }
    
    /** Sets node to n. */
    void set(Node<K,V> n) { node = n; }
    
    /** Sets the entry in this node to (k,v), creating a new node if necessary.  Does NOT adjust size of any associated
      * TreeMap. */
    void set(K k, V v) {
      if (isEmpty()) node = new Node<K,V>(k,v);
      else {
        node.key = k;
        node.value = v;
      }
    }
    
    /** Requires: ! isEmpty(). Sets the key field of node to k. */
    void setKey(K k) {
      assert ! isEmpty();
      node.key = k; 
    }
    
    /** Requires:! isEmpty(). Sets the value field of node to v. */
    void setValue(V v) { 
      assert ! isEmpty();
      node.value = v; 
    }
    
    /** Requires: ! isEmpty(). Sets the left field of node to rn. */
    void setLeft(RefNode<K,V> rn) { 
      assert ! isEmpty();
      node.left = rn; 
    }
    
    /** Requires: ! isEmpty(). Sets the right field of node to rn. */
    void setRight(RefNode<K,V> rn) { 
      assert ! isEmpty();
      node.right = rn;
    }
    
    /** Requires: ! isEmpty().
      * Deletes the entry in this and decrements size. 
      * @param map is the TreeMap containing this RefNode; required to update size.
      * @return the value in the deleted entry; null if this is empty. */ 
    V remove(TreeMap<K,V> map) {
      assert ! isEmpty();
      V oldVal = node.value;
      if (left().isEmpty()) node = right().node;
      else if (right().isEmpty()) node = left().node;  
      else { // both left and right subtrees are non-empty
        Node<K,V> minNode = right().removeMin();
        assert minNode != null;
        node.key = minNode.key;
        node.value = minNode.value;
      }
      (map.size)--;
      return oldVal;
    }
    
    /** Requires: ! isEmpty(). 
      * Deletes the minimum Node in this tree.  Does NOT decrement size of the containing TreeMap.
      * @return the deleted Node. */
    private Node<K,V> removeMin() {
      assert ! isEmpty();
      RefNode<K,V> minRefNode = findFirst();  // return minimum node in this tree
      Node<K,V> minNode = minRefNode.node;
      minRefNode.set(minNode.right.node);  // argument may be null
      return minNode;
    }
    
    /** Requires: ! isEmpty().
      * @returns the first non-empty RefNode in this. */
    private RefNode<K,V> findFirst() {
      assert ! isEmpty();
      RefNode<K,V> current = this;
      do {
        RefNode<K,V> next = current.left();
        if (next.isEmpty()) break;  // current is ref to minimum node so far
        current = next;
      } while (true);
      return current;   
    }
    
    /** Requires: ! root.isEmpty()
      * @returns the last non-empty RefNode in this. */
    private RefNode<K,V> findLast() {
      assert node != null;
      RefNode<K,V> current = this;
      do {
        RefNode<K,V> next = current.right();
        if (next.isEmpty()) break;  // current is ref to maximum node so far
        current = next;
      } while (true);
      return current; 
    }
    
    private static class Node<K extends Comparable<K>, V> {
      /* Fields */
      K key;
      V value;
      RefNode<K,V> left = makeEmpty();
      RefNode<K,V> right = makeEmpty();
      
      /* Constructor */
      Node(K k, V v) {
        key = k;
        value = v;
      }
    }
  }
}
